#include <stdio.h>
#include <stdlib.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>

#define XDIM 640
#define YDIM 512

#define maxiter 200

#include <linux/fb.h>
static class Video
{
    int fd;
    unsigned char *p;
    unsigned xres;
public:
    Video() : xres(1504)
    {
        fd = open("/dev/fb0", O_RDWR);
        
        struct fb_cmap cmap;
        cmap.start=0;
        cmap.len  =256;
        cmap.red  =(__u16 *)calloc(256, 2);
        cmap.green=(__u16 *)calloc(256, 2);
        cmap.blue =(__u16 *)calloc(256, 2);
        cmap.transp=NULL;

        ioctl(fd, FBIOGETCMAP, &cmap);
        for(int a=0; a<120; a++)
        {
            double mag = a/120.0;
            
            int r = int(mag*511);
            int g = int(mag*511)-256;
            int b = int(mag*2400)-(2400-256);
            
            if(r>255)r=255;if(r<0)r=0;
            if(g>255)g=255;if(g<0)g=0;
            if(b>255)b=255;if(b<0)b=0;
            
            cmap.red  [a+16] = r * 65535 / 255;
            cmap.green[a+16] = g * 65535 / 255;
            cmap.blue [a+16] = b * 65535 / 255;
            
            cmap.red  [a+16+120] = b * 65535 / 255;
            cmap.green[a+16+120] = g * 65535 / 255;
            cmap.blue [a+16+120] = r * 65535 / 255;
        }
        ioctl(fd, FBIOPUTCMAP, &cmap);

        free(cmap.red);
        free(cmap.green);
        free(cmap.blue);

        p = (unsigned char *)mmap(NULL, xres*1024, PROT_WRITE, MAP_SHARED, fd, 0);
        if((int)p==-1)
        {
            perror("mmap");
            throw;
        }
    }
    ~Video()
    {
        munmap(p, xres*1024);
    }
    void operator() (unsigned x, unsigned y, unsigned char c)
    {
        /*p[y*xres+x] = c;*/
        x+=(1280-XDIM*2)/4; y+=(1024-YDIM*2)/4;
        p[(y*2  )*xres+(x*2  )] = c;
        p[(y*2  )*xres+(x*2+1)] = c;
        p[(y*2+1)*xres+(x*2  )] = c;
        p[(y*2+1)*xres+(x*2+1)] = c;
    }
} PSet;

static class Fract
{
    static const double XMin = -1.4;
    static const double XMax = 1.4;
    static const double YMin = -1.4;
    static const double YMax = 1.4;

    static const double RMin = -2.02;
    static const double RMax = +0.5;
    static const double IMin = -1.15;
    static const double IMax = +1.15;
    
    inline int juliacalc(double XP,double YP, double RP,double IP)
    {
        register double r = XP, i = YP;
        for(int c=0; ; )
        {
            const double r2 = r*r;
            const double i2 = i*i;
            
            if(r2+i2 >= 4)
            {
                /* Would never return back to <4. Not part of set. Exit. */
                return -1;
            }
            
            i = r*i*2 + IP;
            r = r2-i2 + RP;
            
            if(++c >= maxiter)
            {
                /* Tried enough. We are inside. See how inside. */
                return 1+ int((r2+i2) * (120) / 2);
            }
        }
    }
    inline int mandelcalc(double RP,double IP)
    {
        register double r = RP, i = IP;
        for(int c=0; ; )
        {
            const double r2 = r*r;
            const double i2 = i*i;
            
            if(r2+i2 >= 4)
            {
                /* Would never return back to <4. Not part of set. Exit. */
                return -c*120/maxiter;
            }
            
            i = r*i*2 + IP;
            r = r2-i2 + RP;
            
            if(++c >= maxiter)
            {
                /* drand48() on tss ditherin takia */
                int c = int((r2+i2)*500+drand48());
                if(c<1)c=1;
                if(c>120)c=120;
                /* Tried enough. We are inside. See how inside. */
                return c;
            }
        }
    }
    
public:
    
    int get(unsigned x,unsigned y, unsigned r,unsigned i)
    {
        return juliacalc(
            XMin+x*(XMax-XMin)/XDIM,
            YMin+y*(YMax-YMin)/YDIM,
            RMin+r*(RMax-RMin)/XDIM,
            IMin+i*(IMax-IMin)/YDIM);
    }
    int mandel(unsigned r,unsigned i)
    {
        return mandelcalc(
            RMin+r*(RMax-RMin)/XDIM,
            IMin+i*(IMax-IMin)/YDIM);
    }
} Fract;

int main(void)
{
    bool inside[XDIM][YDIM];
    for(int y=YDIM-1; y>=0; --y)
        for(int x=XDIM-1; x>=0; --x)
        {
            int c = Fract.mandel(x,y);
            if(c > 0)
            {
                inside[x][y] = true;
                PSet(x,y, 120+16-1 -(c-1));
            }
            else
            {
                inside[x][y] = false;
                PSet(x,y, 120+16+ -c);
            }
        }
    
    puts("You can interrupt now...");
    sleep(1);
    puts("Or could...");

    for(int w=0; w<YDIM; w+=30)
//    int w = DIM * 4/8;

    for(int y=0; y<YDIM; ++y)
        for(int x=XDIM-1; x>=0; --x)
        {
            if(!inside[x][y])continue;
            PSet(x,y, 1);
            
            for(int z=0; z<XDIM; ++z)
            {
                int c = Fract.get(z, w, x,y);
                if(c > 0)
                {
                    //c = 0; //255-(z*(256-16)/DIM));
                    PSet(x,y-z*30/XDIM, 16+c);
                    break;
                }
            }
        }
    
    return 0;
}
